"""
Annotation tasks tell annotation sets how to parse annotations.

Remo currently supports the following annotation tasks:

- Object detection
- Instance segmentation
- Image classification

For each of these tasks, Remo supports a variety of annotation formats.
You can [read more about supported annotation formats here](/docs/annotation-formats/)

"""

object_detection = 'Object detection'
instance_segmentation = 'Instance segmentation'
image_classification = 'Image classification'
