# Frontend endpoints

datasets = '/datasets'
annotation = '/annotation/{}'
annotation_set_insights = 'annotation-detail/{}/insights'
search = '/datasets/filtered/images'
image_view = '/image/{}?dataset_id={}'
annotate_image = '/annotation/{}/image/{}'
